"""The AWS vmss Action. This will pull the AWS ASGs"""
from threemystic_cloud_data_client.cloud_providers.aws.client.actions.base_class.base import cloud_data_client_aws_client_action_base as base
import asyncio
from threemystic_cloud_data_client.cloud_providers.aws.client.actions.vmimage import cloud_data_client_aws_client_action as vmimage_data

class cloud_data_client_aws_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="vmss",
      logger_name= "cloud_data_client_aws_client_action_vmss",
      *args, **kwargs)
    
    
    self.data_id_name = "AutoScalingGroupARN"
    
    self.arn_lambda = (lambda item: item["raw_item"][self.data_id_name])
    
    self.auto_region_resourcebytype= ["Amazon Elastic Compute Cloud - Compute"]
    self.resource_group_filter = [
    {
      'Name': 'resource-type',
      'Values': [
        'AWS::EC2::Instance',
      ]
    },
    {
      'Name': 'resource-type',
      'Values': [
        'AWS::AutoScaling::AutoScalingGroup',
      ]
    }
  ]
    
  
  async def _pre_load_main_process(self, pool, *args, **kwarg):
    self.required_extra_data["instances_information"] = {}
    if self.required_extra_data.get("ami") is None:
      base_action_params = {
        "cloud_data_client": self._get_cloud_data_client_raw(),
        "common": self.get_common(),
        "logger": self.get_common().get_logger()
      }
     
      ami_data = vmimage_data(
        **base_action_params
      )
      tmp_ami_data = await ami_data.main(pool=pool, run_params= self.get_runparam_key(data_key= None))
      self.required_extra_data["ami"] = await ami_data.get_data_by_id(results= tmp_ami_data, id_override = "extra_id_only")

  def _process_preloaded_data(self, preloaded_data = None, *args, **kwargs):
    self.required_extra_data = {}
    if preloaded_data is not None:
      if preloaded_data.get("ami") is not None:        
        if len(preloaded_data.get("ami")) > 0:
          self.required_extra_data["ami"] = preloaded_data.get("ami")
  
  async def __getImageInfo(self, image_id, *args, **kwargs):
    if image_id in self.required_extra_data["ami"]:
      return {
        "Id": image_id,
        "Name": self.required_extra_data["ami"][image_id]["Name"],
        "Description": self.required_extra_data["ami"].get("Description")
      }


    return {
      "Id": image_id,
    }
  
  async def process_account_region_asgInstances_get_launchtemplate_version(self, ec2_client, id, version, *args, **kwargs):
      lt_details = self.get_cloud_client().general_boto_call_array(
        boto_call=lambda item: ec2_client.describe_launch_template_versions(**item),
        boto_params={"LaunchTemplateId": id, "Versions": [str(version)] },
        boto_nextkey = None,
        boto_key="LaunchTemplateVersions"
      )

      if len(lt_details) < 1:
          return {}

      return (await self.__get_imagedetails_template(lt_details))[0]
  
  async def process_account_region_asgInstances_get_launchtemplate_getversions(self, ec2_client, id, minversion, maxversion, *args, **kwargs):
      
    minversion = minversion if minversion > 1 else 1
    if minversion == maxversion:
      return [(await self.process_account_region_asgInstances_get_launchtemplate_version(ec2_client, id, minversion))]
    
    lt_details = self.get_cloud_client().general_boto_call_array(
      boto_call=lambda item: ec2_client.describe_launch_template_versions(**item),
      boto_params={"LaunchTemplateId": id, "MinVersion": str(minversion if minversion > 1 else 1), "MaxVersion": str(maxversion)  },
      boto_nextkey = None,
      boto_key="LaunchTemplateVersions"
    )

    if len(lt_details) < 1:
      return []

    return await self.__get_imagedetails_template(lt_details)
  
  async def process_account_region_asgInstances_get_launchtemplate_getversionlast5(self, ec2_client, id, latest_version, *args, **kwargs):
      last10_versions = await self.process_account_region_asgInstances_get_launchtemplate_getversions(ec2_client, id, latest_version - 10, latest_version)
      if len(last10_versions) < 2:
        return last10_versions
        
      last10_versions.sort(key=lambda item: item["VersionNumber"], reverse=True)

      if len(last10_versions) < 5:
        return last10_versions
      
      return last10_versions[:5]

  async def __get_imagedetails_template(self, templates_details ):
    for lt_details in templates_details:
      lt_details["LaunchTemplateData"]["extra_image"] = await self.__getImageInfo(image_id= lt_details["LaunchTemplateData"]["ImageId"])
    
    return templates_details
  
  async def process_account_region_asgInstances_get_launchtemplates(self, ec2_client, launch_tempalte_details, loop, *args, **kwargs):
      lt_details = {}
      launch_template_ids = []

      for version, template_ids in launch_tempalte_details.items():
          for id in template_ids:
            if lt_details.get(id) is None:
              launch_template_ids.append(id)
              lt_details[id] = {
                "versions": {
                  version.lower(): await self.process_account_region_asgInstances_get_launchtemplate_version(ec2_client, id, version) 
                }
              }           
              continue
              
            if version in lt_details[id]["versions"]:
              continue
            
            lt_details[id]["versions"][version] = await self.process_account_region_asgInstances_get_launchtemplate_version(ec2_client, id, version) 
      
      launch_templates_details = { item["LaunchTemplateId"]:item for item in self.get_cloud_client().general_boto_call_array(
            boto_call=lambda item: ec2_client.describe_launch_templates(**item),
            boto_params={"LaunchTemplateIds": launch_template_ids },
            boto_nextkey = "NextToken",
            boto_key="LaunchTemplates"
        )}
      
      get_last5_tasks = []
      for id, details in lt_details.items():
        details["template"] = launch_templates_details[id]
        if "$latest" not in details["versions"]:
          details["versions"]["$latest"] = (await self.process_account_region_asgInstances_get_launchtemplate_version(ec2_client, id, "$Latest"))
          

        if "$default" not in details["versions"]:
          details["versions"]["$default"] = (await self.process_account_region_asgInstances_get_launchtemplate_version(ec2_client, id, "$Default"))

        
        latest_version = details["versions"]["$latest"]["VersionNumber"]
        default_version = details["versions"]["$default"]["VersionNumber"]

        if default_version not in details["versions"]:
          details["versions"][default_version] = details["versions"]["$default"]
        try:
          if latest_version not in details["versions"]:
            details["versions"][latest_version] = details["versions"]["$latest"]
        except:
            print(latest_version)
            print(details)
            raise

        get_last5_tasks.append(loop.create_task(self.process_account_region_asgInstances_get_launchtemplate_getversionlast5(ec2_client=ec2_client, id=id, latest_version=latest_version)))
        
      if len(get_last5_tasks) < 1:
        return lt_details
        
      await asyncio.wait(get_last5_tasks)
      for task in get_last5_tasks:
        versions = task.result()
        for version in versions:
          if version["VersionNumber"] in lt_details[version["LaunchTemplateId"]]["versions"]:
            continue

          lt_details[version["LaunchTemplateId"]]["versions"][version["VersionNumber"]] = version
      
      return lt_details
  
  async def process_account_region_asgInstances_get_launchconfigs(self, auto_scaling_client, launch_configuration_names = None):
    boto_params = {}

    if launch_configuration_names is not None and len(launch_configuration_names) < 1:
      return {}
        
    if launch_configuration_names is not None:
      boto_params["LaunchConfigurationNames"] = launch_configuration_names
    
    return {config["LaunchConfigurationName"]:config for config in  self.get_cloud_client().general_boto_call_array(
      boto_call=lambda item: auto_scaling_client.describe_launch_configurations(**item),
      boto_params= boto_params,
      boto_nextkey = "NextToken",
      boto_key="LaunchConfigurations"
    ) } 
  
  def __get_launch_template_version(self, launch_template_details, *args, **kwarg):
    return launch_template_details["Version"] if launch_template_details.get("Version") is not None else "$Default"
  
  async def _process_account_data_region(self, account, region, resource_groups, loop, *args, **kwargs):
    auto_scaling_client = self.get_cloud_client().get_boto_client(
      client= "autoscaling",
      account= account,
      region= region
    )

    ec2_client = self.get_cloud_client().get_boto_client(
      client= "ec2",
      account= account,
      region= region
    )

    groups = self.get_cloud_client().general_boto_call_array(
      boto_call=lambda item: auto_scaling_client.describe_auto_scaling_groups(**item),
      boto_params={ },
      boto_nextkey = "NextToken",
      boto_key="AutoScalingGroups"
    )

    launch_configuration_names = []
    launch_tempalte_details = {}
    for g in groups:
      if 'LaunchConfigurationName' in g:
        launch_configuration_names.append(g['LaunchConfigurationName'])
        continue
      
      if 'MixedInstancesPolicy' in g and 'LaunchTemplate' not in g:
        g['LaunchTemplate'] = g['MixedInstancesPolicy']['LaunchTemplate']['LaunchTemplateSpecification']

      if 'LaunchTemplate' in g:
        template_version = self.__get_launch_template_version(g['LaunchTemplate'])
        if launch_tempalte_details.get(template_version) is None:
          launch_tempalte_details[template_version] = []
        launch_tempalte_details[template_version].append(g['LaunchTemplate']["LaunchTemplateId"])
        continue

        

    task_deployconfig = {
      "config": loop.create_task(self.process_account_region_asgInstances_get_launchconfigs(auto_scaling_client, launch_configuration_names)),
      "template": loop.create_task(self.process_account_region_asgInstances_get_launchtemplates(ec2_client, launch_tempalte_details, loop))
    }

    await asyncio.wait(task_deployconfig.values())

    return {
      "region": region,
      "resource_groups": resource_groups,
      "data": [
          self.get_common().helper_type().dictionary().merge_dictionary([
            {},
            {
              "extra_launchconfig": self.get_common().helper_type().dictionary().merge_dictionary([
                {},
                {"extra_image": await self.__getImageInfo(image_id= task_deployconfig["config"].result().get(item['LaunchConfigurationName'])["ImageId"])},
                task_deployconfig["config"].result().get(item['LaunchConfigurationName'])
              ]) if item.get('LaunchConfigurationName') is not None else {},
              "extra_launchtemplate": task_deployconfig["template"].result().get(item['LaunchTemplate']["LaunchTemplateId"]) if item.get('LaunchTemplate') is not None else {},
            },
            item
          ]) for item in groups
        ]
    }



from threemystic_cloud_data_client.cloud_providers.azure.client.actions.base_class.base import cloud_data_client_azure_client_action_base as base
import asyncio
from requests import post as request_post
from decimal import Decimal, ROUND_HALF_UP
from azure.mgmt.costmanagement import CostManagementClient
from azure.mgmt.costmanagement.models import GranularityType,ForecastDefinition,ForecastType,ForecastTimeframe,ForecastTimePeriod,QueryDefinition,TimeframeType,ExportType,QueryTimePeriod,QueryResult

class cloud_data_client_azure_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="budget", 
      logger_name= "cloud_data_client_azure_client_action_budget",
      *args, **kwargs)
  
  def __process_get_cost_generate_data(self, account, client, cost_filter, is_forcast = False, *args, **kwargs):
    if not is_forcast:
      cost_filter = QueryDefinition(**cost_filter)
      cost_response = self.get_cloud_client().sdk_request(
        tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
        lambda_sdk_command=lambda: client.query.usage(
          scope= f'{self.get_cloud_client().get_account_prefix()}{self.get_cloud_client().get_account_id(account= account)}',
          parameters= cost_filter
        )
      )
    else:
      cost_filter = ForecastDefinition(**cost_filter)
      cost_response = self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True),
          lambda_sdk_command=lambda: client.forecast.usage(
            scope= f'{self.get_cloud_client().get_account_prefix()}{self.get_cloud_client().get_account_id(account= account)}',
            parameters= ForecastDefinition(**cost_filter),
          )
        )
    
    
    next_link = getattr(cost_response, "next_link", None)
    
    next_cost_page = 1
    cost_filter = cost_filter.serialize()
    if self.get_common().helper_type().general().is_type(obj= cost_filter, type_check= dict):
      cost_filter = self.get_common().helper_json().dumps(data= cost_filter)

    while not self.get_common().helper_type().string().is_null_or_whitespace(string_value= next_link):
      cost_api_resquest = request_post(
        url= next_link,
        headers= {
          "Authorization": f"Bearer {self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)).get_token('https://management.azure.com/.default').token}",
          "Content-Type": "application/json"
        },
        data= cost_filter,
        timeout= 10,
      )

      if cost_api_resquest.status_code == 429:
        self.get_common().helper_type().requests().expodential_backoff_wait(attempt= next_cost_page, auto_sleep= True)
        next_cost_page += 1
        continue

      cost_api_response = cost_api_resquest.json()
      next_link = cost_api_response["properties"].get("nextLink")
      next_cost_page = 1

      cost_response.rows += cost_api_response["properties"].get("rows")

    return cost_response
   
  async def __process_get_cost_data_forcast_time_range(self, account, year_data, cost_metric, start_date, end_date, fiscal_start, fiscal_end, query_grouping = [], *args, **kwargs):
    
    cost_filter = {
      'type': getattr(ForecastType, cost_metric),
      'timeframe': ForecastTimeframe.CUSTOM,
      'time_period': ForecastTimePeriod(
        from_property= self.get_common().helper_type().datetime().parse_iso(iso_datetime_str= f"{self.get_common().helper_type().datetime().datetime_as_string(dt_format='%Y-%m-%d', dt= start_date)}T00:00:00+00:00"),
        to= self.get_common().helper_type().datetime().parse_iso(iso_datetime_str= f"{self.get_common().helper_type().datetime().datetime_as_string(dt_format='%Y-%m-%d', dt= end_date)}T23:59:59+00:00")
      ),
      'dataset': {
        'granularity': GranularityType.DAILY,
        'aggregation': {
          f'totalCost{self.get_common().helper_type().string().set_case(self.get_cloud_data_client().get_default_currency(), case= "upper")}': {
            'name': f'Cost{self.get_common().helper_type().string().set_case(self.get_cloud_data_client().get_default_currency(), case= "upper")}',
            'function': 'Sum'
          }
        },
        'grouping': [ 
          {"type": "Dimension", "name": dimension} for dimension in query_grouping
        ],
      }
    }

    try:
      usage = self.__process_get_cost_generate_data(account= account, cost_filter= cost_filter, is_forcast= True, *args, **kwargs)

      if usage is None:
        return usage

      return await self.__process_get_cost_data_daily_data(usage= usage, year_data= year_data, cost_metric= cost_metric, fiscal_start= fiscal_start, fiscal_end= fiscal_end,  is_forcast= True, query_grouping= query_grouping)
    except Exception as err:
      self.get_common().get_logger().exception(msg= f"{self.get_cloud_client().get_account_id(account= account)} - {str(err)}", extra={"exception": err})
      return {}

    
  async def __process_get_cost_data_time_range(self, account, year_data, cost_metric, start_date, end_date, fiscal_start, fiscal_end, query_grouping = [], *args, **kwargs):
    
    cost_filter = {
      'type': getattr(ExportType, cost_metric),
      'timeframe': TimeframeType.CUSTOM,
      'time_period': QueryTimePeriod(
        from_property= self.get_common().helper_type().datetime().parse_iso(iso_datetime_str= f"{self.get_common().helper_type().datetime().datetime_as_string(dt_format='%Y-%m-%d', dt= start_date)}T00:00:00+00:00"),
        to=  self.get_common().helper_type().datetime().parse_iso(iso_datetime_str= f"{self.get_common().helper_type().datetime().datetime_as_string(dt_format='%Y-%m-%d', dt= end_date)}T23:59:59+00:00")
      ),
      'dataset': {
        'granularity': GranularityType.DAILY,
        'aggregation': {
          f'totalCost': {
            'name': f'Cost',
            'function': 'Sum'
          }
        },
        'grouping': [ 
          {"type": "Dimension", "name": dimension} for dimension in query_grouping
        ]        
      }
    }
    
    try:
      usage = self.__process_get_cost_generate_data(account= account, cost_filter= cost_filter, is_forcast= False, *args, **kwargs)
      
      if usage is None:
        return {}

      return await self.__process_get_cost_data_daily_data(account= account, year_data= year_data, cost_metric= cost_metric, usage= usage, fiscal_start= fiscal_start, fiscal_end= fiscal_end, is_forcast= False, query_grouping= query_grouping)
    except Exception as err:
      self.get_common().get_logger().exception(msg= f"{self.get_cloud_client().get_account_id(account= account)} - {str(err)}", extra={"exception": err})
      return {}
    
  def __init_costdata_month(self, data_dt, *args, **kwargs):
    return {
      "currency": self.get_common().helper_type().string().set_case(self.get_cloud_data_client().get_default_currency(), case= "upper"),
      "month": data_dt.month,
      "year": data_dt.year,
      "totals":{
        "total": Decimal(0),
        "fiscal_total": Decimal(0),
        "forcast_total": Decimal(0),
        "fiscal_forcast_total": Decimal(0),
        "resource_group": {
          "total":{},
          "forcast_total":{},
          "origional_currency_total":{},
          "origional_currency_forcast_total":{},
        },
        "resource_type": {
          "total":{},
          "forcast_total":{},
          "origional_currency_total":{},
          "origional_currency_forcast_total":{},
        }
      },
      "days":{}
    }
  
  def __init_costdata_month_day(self, data_dt, currency, *args, **kwargs):
    return {
      "currency": self.get_common().helper_type().string().set_case(self.get_cloud_data_client().get_default_currency(), case= "upper"),
      "origional_currency": self.get_common().helper_type().string().set_case(string_value= currency, case= "upper"),
      "date": data_dt,
      "total": Decimal(0),
      "forcast_total": Decimal(0),
      "origional_currency_total": Decimal(0),
      "origional_currency_forcast_total": Decimal(0),
      "resource_group": {
        "total":{},
        "forcast_total":{},
        "origional_currency_total":{},
        "origional_currency_forcast_total":{},
      },
      "resource_type": {
        "total":{},
        "forcast_total":{},
        "origional_currency_total":{},
        "origional_currency_forcast_total":{},
      }
    }
    
  async def __process_get_cost_data_daily_data(self, usage, year_data, cost_metric, fiscal_start, fiscal_end, query_grouping, is_forcast = False, *args, **kwargs):
    
    if year_data is None:
      year_data = {}
    if year_data.get(cost_metric) is None:
      year_data[cost_metric] = {}

    column_indexs = {
      self.get_common().helper_type().string().set_case(string_value= dimension, case= "lower"):-1 for dimension in query_grouping
    }
    if column_indexs.get("resourcegroup") is None:
      column_indexs["resourcegroup"] = -1
    if column_indexs.get("resourcetype") is None:
      column_indexs["resourcetype"] = -1

    column_indexs["cost"] = -1
    column_indexs[f"cost{self.get_common().helper_type().string().set_case(self.get_cloud_data_client().get_default_currency(), case= 'lower')}"] = -1
    column_indexs["usagedate"] = -1
    column_indexs["currency"] = -1

    total_key = "total" if not is_forcast else "forcast_total"
    

    for index, data in enumerate(usage.columns):
      
      if column_indexs.get(self.get_common().helper_type().string().set_case(string_value= data.name , case= "lower")) is None:
        continue
      column_indexs[self.get_common().helper_type().string().set_case(string_value= data.name, case= "lower")] = index
    
    cost_key = "cost"

    if column_indexs[cost_key] < 0:
      cost_key = f"cost{self.get_common().helper_type().string().set_case(self.get_cloud_data_client().get_default_currency(), case= 'lower')}"

    for cost_data in usage.rows:
      data_dt = self.get_common().helper_type().datetime().datetime_from_string(dt_string= str(cost_data[column_indexs["usagedate"]]), dt_format= "%Y%m%d")
      by_month_key = self.get_common().helper_type().datetime().datetime_as_string(dt_format= "%Y%m", dt= data_dt)
        
      day_key = self.get_common().helper_type().datetime().datetime_as_string(dt_format= "%Y%m%d", dt= data_dt)
      if year_data[cost_metric].get(by_month_key) is None:
        year_data[cost_metric][by_month_key] = self.__init_costdata_month(data_dt= data_dt)
      
      if year_data[cost_metric][by_month_key]["days"].get(day_key) is None:
        year_data[cost_metric][by_month_key]["days"][day_key] = self.__init_costdata_month_day(data_dt= data_dt, currency= cost_data[column_indexs["currency"]])

      
      raw_row_data_cost = (cost_data[column_indexs[cost_key]])
      row_data_cost = (cost_data[column_indexs[cost_key]])
      
      if year_data[cost_metric][by_month_key]["days"][day_key]["currency"] != year_data[cost_metric][by_month_key]["days"][day_key]["origional_currency"]:
        row_data_cost = self.get_common().helper_currency().convert(
          ammount= row_data_cost,
          currency_from= year_data[cost_metric][by_month_key]["days"][day_key]["origional_currency"],
          currency_to= year_data[cost_metric][by_month_key]["days"][day_key]["currency"],
          conversion_date= self.get_common().helper_type().datetime().yesterday(dt=self.get_common().helper_type().datetime().datetime_from_string(
            dt_string= self.get_common().helper_type().datetime().datetime_as_string(
              dt= data_dt,
              dt_format= "%Y%m01"
            ),
            dt_format= "%Y%m%d"
          )).date()
        )

      year_data[cost_metric][by_month_key]["days"][day_key][f'origional_currency_{total_key}'] += Decimal(raw_row_data_cost)
      year_data[cost_metric][by_month_key]["days"][day_key][f'{total_key}'] += Decimal(row_data_cost)
      year_data[cost_metric][by_month_key]["totals"][f'{total_key}'] += Decimal(row_data_cost)
      if data_dt >= fiscal_start and data_dt <= fiscal_end:
        year_data[cost_metric][by_month_key]["totals"][f'fiscal_{total_key}'] += Decimal(row_data_cost)

      if column_indexs["resourcegroup"] > -1:
        if (year_data[cost_metric][by_month_key]["days"][day_key]["resource_group"][f'{total_key}'].get(cost_data[column_indexs["resourcegroup"]]) is None or
            year_data[cost_metric][by_month_key]["days"][day_key]["resource_group"][f'origional_currency_{total_key}'].get(cost_data[column_indexs["resourcegroup"]]) is None):
          year_data[cost_metric][by_month_key]["days"][day_key]["resource_group"][f'origional_currency_{total_key}'][cost_data[column_indexs["resourcegroup"]]] = Decimal(0)
          year_data[cost_metric][by_month_key]["days"][day_key]["resource_group"][f'{total_key}'][cost_data[column_indexs["resourcegroup"]]] = Decimal(0)
        
        if (year_data[cost_metric][by_month_key]["totals"]["resource_group"][f'{total_key}'].get(cost_data[column_indexs["resourcegroup"]]) is None or
            year_data[cost_metric][by_month_key]["totals"]["resource_group"][f'origional_currency_{total_key}'].get(cost_data[column_indexs["resourcegroup"]]) is None):
          year_data[cost_metric][by_month_key]["totals"]["resource_group"][f'origional_currency_{total_key}'][cost_data[column_indexs["resourcegroup"]]] = Decimal(0)
          year_data[cost_metric][by_month_key]["totals"]["resource_group"][f'{total_key}'][cost_data[column_indexs["resourcegroup"]]] = Decimal(0)

        year_data[cost_metric][by_month_key]["days"][day_key]["resource_group"][f'origional_currency_{total_key}'][cost_data[column_indexs["resourcegroup"]]] += Decimal(raw_row_data_cost)
        year_data[cost_metric][by_month_key]["days"][day_key]["resource_group"][f'{total_key}'][cost_data[column_indexs["resourcegroup"]]] += Decimal(row_data_cost)
        year_data[cost_metric][by_month_key]["totals"]["resource_group"][f'origional_currency_{total_key}'][cost_data[column_indexs["resourcegroup"]]] += Decimal(raw_row_data_cost)
        year_data[cost_metric][by_month_key]["totals"]["resource_group"][f'{total_key}'][cost_data[column_indexs["resourcegroup"]]] += Decimal(row_data_cost)

      if column_indexs["resourcetype"] > -1:
        if (year_data[cost_metric][by_month_key]["days"][day_key]["resource_type"][f'{total_key}'].get(cost_data[column_indexs["resourcetype"]]) is None or
            year_data[cost_metric][by_month_key]["days"][day_key]["resource_type"][f'origional_currency_{total_key}'].get(cost_data[column_indexs["resourcetype"]]) is None):
          year_data[cost_metric][by_month_key]["days"][day_key]["resource_type"][f'origional_currency_{total_key}'][cost_data[column_indexs["resourcetype"]]] = Decimal(0)
          year_data[cost_metric][by_month_key]["days"][day_key]["resource_type"][f'{total_key}'][cost_data[column_indexs["resourcetype"]]] = Decimal(0)
        
        if (year_data[cost_metric][by_month_key]["totals"]["resource_type"][f'{total_key}'].get(cost_data[column_indexs["resourcetype"]]) is None or
            year_data[cost_metric][by_month_key]["totals"]["resource_type"][f'origional_currency_{total_key}'].get(cost_data[column_indexs["resourcetype"]]) is None):
          year_data[cost_metric][by_month_key]["totals"]["resource_type"][f'origional_currency_{total_key}'][cost_data[column_indexs["resourcetype"]]] = Decimal(0)
          year_data[cost_metric][by_month_key]["totals"]["resource_type"][f'{total_key}'][cost_data[column_indexs["resourcetype"]]] = Decimal(0)


        year_data[cost_metric][by_month_key]["days"][day_key]["resource_type"][f'origional_currency_{total_key}'][cost_data[column_indexs["resourcetype"]]] += Decimal(raw_row_data_cost)
        year_data[cost_metric][by_month_key]["days"][day_key]["resource_type"][f'{total_key}'][cost_data[column_indexs["resourcetype"]]] += Decimal(row_data_cost)
        year_data[cost_metric][by_month_key]["totals"]["resource_type"][f'origional_currency_{total_key}'][cost_data[column_indexs["resourcetype"]]] += Decimal(raw_row_data_cost)
        year_data[cost_metric][by_month_key]["totals"]["resource_type"][f'{total_key}'][cost_data[column_indexs["resourcetype"]]] += Decimal(row_data_cost)

  async def __process_get_cost_data_sum_year_forcast_data(self, year_month_data, forcast_month_data, *args, **kwargs):
    key_list = self.get_common().helper_type().list().unique_list(data= list(year_month_data.keys()) + list(forcast_month_data.keys()))

    for key in key_list:
      year_month_key_data = year_month_data.get(key)
      forcast_month_key_data = forcast_month_data.get(key)

      if (self.get_common().helper_type().general().is_type(obj=year_month_key_data, type_check= dict) or
        self.get_common().helper_type().general().is_type(obj=forcast_month_key_data, type_check= dict)):
        return await self.__process_get_cost_data_sum_year_forcast_data(
          year_month_data= year_month_key_data,
          forcast_month_data= forcast_month_key_data
        )
      
      if self.get_common().helper_type().general().is_numeric(year_month_key_data) and forcast_month_key_data is None:
        forcast_month_key_data = Decimal(0)
      
      if self.get_common().helper_type().general().is_numeric(year_month_key_data) and year_month_key_data is None:
        year_month_key_data = Decimal(0)
      
      year_month_data[key] = (year_month_key_data + forcast_month_key_data)
        

  async def __process_get_cost_data_process_forcast_year_data(self, year_data, cost_metric, start_date, end_date, fiscal_start, fiscal_end, *args, **kwargs):
     end_date += self.get_common().helper_type().datetime().time_delta(months= 1, dt= end_date)
     while start_date < end_date:
      await self.__process_get_cost_data_forcast_time_range(
          year_data= year_data,
          cost_metric= cost_metric,
          start_date= start_date,
          end_date= self.get_common().helper_type().datetime().yesterday(dt=(start_date + self.get_common().helper_type().datetime().time_delta(months= 3, dt= start_date))),
          fiscal_start= fiscal_start, fiscal_end= fiscal_end, 
          query_grouping= ["SubscriptionId"],
          *args, **kwargs )

      start_date = start_date + self.get_common().helper_type().datetime().time_delta(months= 3, dt= start_date)

  async def __process_get_cost_data_process_year_data(self, year_data, cost_metric, start_date, end_date, fiscal_start, fiscal_end, *args, **kwargs):
     end_date += self.get_common().helper_type().datetime().time_delta(months= 1, dt= end_date)
     while start_date < end_date and start_date <  self.get_data_start():
      adjusted_enddate = self.get_common().helper_type().datetime().yesterday(dt=(start_date + self.get_common().helper_type().datetime().time_delta(months= 3, dt= start_date)))
      if adjusted_enddate > self.get_data_start():
        adjusted_enddate = self.get_data_start()

      await self.__process_get_cost_data_time_range(
        year_data= year_data,
        cost_metric= cost_metric,
        start_date= start_date,
        end_date= adjusted_enddate,
        fiscal_start= fiscal_start, fiscal_end= fiscal_end, 
        query_grouping= ["SubscriptionId"],
        *args, **kwargs )
      

      start_date = start_date + self.get_common().helper_type().datetime().time_delta(months= 3, dt= start_date)

      

  async def __process_get_cost_data(self, loop, fiscal_year_start, *args, **kwargs):
    fiscal_year_start_date = self.get_common().helper_type().datetime().datetime_from_string(
      dt_string= f"{self.get_data_start().year}/{fiscal_year_start}",
      dt_format= "%Y/%m/%d"
    )
    
    if fiscal_year_start_date > self.get_data_start():
      fiscal_year_start_date = fiscal_year_start_date + self.get_common().helper_type().datetime().time_delta(years= -1)

    
    fiscal_year_end = self.get_common().helper_type().datetime().yesterday(dt= (fiscal_year_start_date
                 + self.get_common().helper_type().datetime().time_delta(years= 1, dt= fiscal_year_start_date)))
    
    start_date = (fiscal_year_start_date
                 + self.get_common().helper_type().datetime().time_delta(months= -1, dt= fiscal_year_start_date))
    
    forecast_end = (fiscal_year_end
                 + self.get_common().helper_type().datetime().time_delta(months= 3, dt= fiscal_year_end))
    year_data = {}
    cost_metric = "AMORTIZED_COST"

    await self.__process_get_cost_data_process_year_data(
      year_data= year_data,
      cost_metric= cost_metric,
      start_date= start_date,
      end_date= self.get_data_start(),
      fiscal_start= fiscal_year_start_date,
      fiscal_end= fiscal_year_end, *args, **kwargs
    )

    
    await self.__process_get_cost_data_process_forcast_year_data(
      year_data= year_data,
      cost_metric= cost_metric,
      start_date= self.get_data_start(),
      end_date= forecast_end,
      fiscal_start= fiscal_year_start_date,
      fiscal_end= fiscal_year_end, *args, **kwargs
    )
    
    month_key = self.get_common().helper_type().datetime().datetime_as_string(
      dt= self.get_data_start(),
      dt_format= "%Y%m"
    )
    last_month_key = self.get_common().helper_type().datetime().datetime_as_string(
      dt= (self.get_data_start() - self.get_common().helper_type().datetime().time_delta(days= (self.get_data_start().day + 1))),
      dt_format= "%Y%m"
    )
    return_data = {
      "cost_metric": cost_metric,
      "year_to_date": Decimal(0),  
      "year_forecast": Decimal(0),
      "fiscal_year_to_date": Decimal(0),  
      "fiscal_year_forecast": Decimal(0),
      "month_to_date": Decimal(0),  
      "month_forecast": Decimal(0),
      "last_seven_days": Decimal(0),
      "raw_last_14_days": {},
      "last_month": Decimal(0),
    }

    last14_days = {}
    await self.__process_get_cost_data_time_range(
      year_data= last14_days,
      cost_metric= cost_metric,
      start_date= (self.get_data_start() + self.get_common().helper_type().datetime().time_delta(days= -14)),
      end_date= self.get_data_start(),
      fiscal_start= fiscal_year_start_date, fiscal_end= fiscal_year_start_date, 
      query_grouping= ["SubscriptionId", "ResourceGroup", "ResourceType"],
      *args, **kwargs )

    day_count = 0
    if last14_days.get(cost_metric) is not None:

      for i in range(0,9):
        if day_count >= 7:
          break
        
        month_key_last14 = self.get_common().helper_type().datetime().datetime_as_string(
          dt= (self.get_data_start() - self.get_common().helper_type().datetime().time_delta(days= i)),
          dt_format= "%Y%m"
        )
        day_key = self.get_common().helper_type().datetime().datetime_as_string(
          dt= (self.get_data_start() - self.get_common().helper_type().datetime().time_delta(days= i)),
          dt_format= "%Y%m%d"
        )
        
        if last14_days[cost_metric].get(month_key_last14) is None:
          continue
        if last14_days[cost_metric][month_key_last14]["days"].get(day_key) is None:
          continue
        
        day_count += 1
        return_data["last_seven_days"] += last14_days[cost_metric][month_key_last14]["days"][day_key]["total"]

      return_data["raw_last_14_days"] = {}
      for month_key in last14_days[cost_metric].keys():
        for day_key, day_data in last14_days[cost_metric][month_key]["days"].items():
          return_data["raw_last_14_days"][day_key]= day_data
    

    
    if year_data.get(cost_metric) is not None:
      for data in year_data.get(cost_metric).values():
        return_data["fiscal_year_to_date"] += data["totals"].get("fiscal_total")
        return_data["fiscal_year_forecast"] += (data["totals"].get("fiscal_total") + data["totals"].get("fiscal_forcast_total"))
        if data["year"] == self.get_data_start().year:
          return_data["year_to_date"] += data["totals"].get("total")
          return_data["year_forecast"] += (data["totals"].get("total") + data["totals"].get("forcast_total"))

      
      if year_data[cost_metric].get(month_key) is not None:
        return_data["month_to_date"] = year_data[cost_metric][month_key]["totals"]["total"]
        return_data["month_forecast"] = year_data[cost_metric][month_key]["totals"]["total"] + year_data[cost_metric][month_key]["totals"]["forcast_total"]
      
      if year_data[cost_metric].get(last_month_key) is not None:
        return_data["last_month"] = year_data[cost_metric][last_month_key]["totals"]["total"]


    return return_data

  async def _process_account_data(self, account, loop, *args, **kwargs):
    
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= kwargs.get("fiscal_year_start")):
      kwargs["fiscal_year_start"] = self.get_cloud_data_client().get_default_fiscal_year_start()
    
    costmanagement_client = CostManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    
    return {
      "account": account,
      "data": [ self.get_common().helper_type().dictionary().merge_dictionary([
        {},
        await self.get_base_return_data(
          account= self.get_cloud_client().serialize_resource(resource= account),
          resource_id =  f'{self.get_cloud_client().get_account_prefix()}{self.get_cloud_client().get_account_id(account= account)}',
        ),
        await self.__process_get_cost_data(account= account, client= costmanagement_client, loop= loop, *args, **kwargs)
      ])]
    }


from threemystic_cloud_data_client.cloud_providers.azure.client.actions.base_class.base import cloud_data_client_azure_client_action_base as base
import asyncio
from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.costmanagement.models import TimeframeType, OperatorType, QueryColumnType, QueryDefinition, QueryTimePeriod, QueryDataset, QueryAggregation, QueryGrouping, QueryFilter, QueryComparisonExpression
from azure.mgmt.compute.models import VirtualMachineScaleSet, VirtualMachineScaleSetVM, OSDisk, DataDisk
from azure.mgmt.costmanagement import CostManagementClient

from threemystic_cloud_data_client.cloud_providers.azure.client.actions.vm import cloud_data_client_azure_client_action as vm_action
from threemystic_cloud_data_client.cloud_providers.azure.client.actions.vmss import cloud_data_client_azure_client_action as vmss_action

class cloud_data_client_azure_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="storage", 
      logger_name= "cloud_data_client_azure_client_action_storage",
      *args, **kwargs)
  
  
    
  def __get_cost_by_resource_query_definition_mtd(self):
    mtd = self.get_cloud_client().get_default_querydefinition()
    mtd.timeframe = TimeframeType.MONTH_TO_DATE

    mtd.dataset.filter = QueryFilter(
        dimensions= QueryComparisonExpression(
        name= "ResourceType",
        operator= OperatorType.IN_ENUM,
        values= ["Microsoft.Compute/disks"]
      )
    )
    mtd.dataset.grouping= [
      QueryGrouping(
        type= QueryColumnType.DIMENSION,
        name= "ResourceId"
      )
    ]

    return mtd
  
  def __get_cost_by_resource_query_definition_ytd(self):
    ytd = self.__get_cost_by_resource_query_definition_mtd()
    ytd.timeframe = TimeframeType.CUSTOM
    ytd.time_period = QueryTimePeriod(
      from_property= self.get_common().helper_type().datetime().parse_iso(iso_datetime_str= f"{self.get_data_start().year}-01-01T00:00:00+00:00"),
      to=  self.get_common().helper_type().datetime().parse_iso(iso_datetime_str= f"{self.get_data_start().year}-12-31T23:59:59+00:00")      
    )

    ytd.dataset.filter = QueryFilter(
        dimensions= QueryComparisonExpression(
        name= "ResourceType",
        operator= OperatorType.IN_ENUM,
        values= ["Microsoft.Compute/disks"]
      )
    )
    ytd.dataset.grouping= [
      QueryGrouping(
        type= QueryColumnType.DIMENSION,
        name= "ResourceId"
      )
    ]

    return ytd

  async def _pre_load_main_process(self, pool, *args, **kwarg):
    self.__vm = None
    base_action_params = {
      "cloud_data_client": self._get_cloud_data_client_raw(),
      "common": self.get_common(),
      "logger": self.get_common().get_logger()
    }
    # This is required because #the normal disk command only looks at long lived disks
    vmss_data = vmss_action(
      **base_action_params
    )
    self.__vmss = await vmss_data.main(pool=pool, run_params= self.get_runparam_key(data_key= None)) 

    self.__cost_by_resource_query_definition = {
      "MTD": self.__get_cost_by_resource_query_definition_mtd(),
      "YTD": self.__get_cost_by_resource_query_definition_ytd()
    }

    vm_data = vm_action(
      **base_action_params
    )
      
    self.__vm = await vm_data.main(pool=pool, run_params= self.get_runparam_key(data_key= None))
      

  async def __get_attached_devices_vm_check_os(self, disk, vm_os_disk = None, *args, **kwarg):
    if vm_os_disk is None:
      return False
    
    if vm_os_disk.get("managed_disk") is None:
      return False
    
    return True if vm_os_disk.get("managed_disk")["id"] == disk.id else False

  async def __get_attached_devices_vm_check_data(self, disk, vm_data_disks = None, *args, **kwarg):
      if vm_data_disks is None:
        return False
      
      if len(vm_data_disks) < 1:
        return False
      
      for mdisk in vm_data_disks:
        if mdisk["id"] == disk.id:
          return True 

      return False
  async def __get_attached_devices_vm_check(self, disk, vm, vmss= None, *args, **kwarg):
    if vm.get("storage_profile") is None:
      return None

    tasks = {
      "os_disk": self.__get_attached_devices_vm_check_os(disk= disk, vm_os_disk= vm["storage_profile"].get("os_disk")),
      "data_disks": self.__get_attached_devices_vm_check_data(disk= disk, vm_os_disk= vm["storage_profile"].get("data_disks"))
    }

    await asyncio.wait(tasks.values())

    for _, task in tasks.items():
      if task.result():
        return self.get_common().helper_type().dictionary().merge_dictionary([
          {},
          {
          "extra_disk_group_type": "vm" if vmss is None else "vmss", 
          "extra_disk_group": None if vmss is None else vmss["name"],
          "extra_disk_group_primary": None if vmss is None else vmss["extra_vmss_vms"][0]["id"] == vm["id"] 
          }, 
          vm])

    return None



  async def __get_attached_devices_vm(self, account, disk, loop, *args, **kwarg):
    if self.__vm is None:
      return None
    
    account_id = self.get_cloud_client().get_account_id(account= account)
    if self.__vm.get(account_id) is None:
      return None

    tasks = []
    for vm in self.__vm[account_id]:
      tasks.append(loop.create_task(self.__get_attached_devices_vm_check(vm= vm, disk= disk )))
      done_tasks, tasks = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED, timeout= 1)
      tasks = [task for task in tasks]
      for done in done_tasks:
        if done.result() is not None:
          return [ done.result() ]
    
    return None

  async def __get_attached_devices_vmss(self, account, disk, loop, *args, **kwarg):
    if self.__vmss is None:
      return None

    account_id = self.get_cloud_client().get_account_id(account= account)
    if self.__vmss.get(account_id) is None:
      return None
    
    tasks = []
    for vmss in self.__vmss[account_id]:
      for vm in vmss["extra_vmss_vms"]:
        tasks.append(loop.create_task(self.__get_attached_devices_vm_check(vm= vm, disk= disk, vmss= vmss )))
        done_tasks, tasks = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED, timeout= 1)
        tasks = [task for task in tasks]
        for done in done_tasks:
          if done.result() is not None:
            return [ done.result() ]
    
    return None
      
  async def __get_attached_devices(self, account, disk, loop, *args, **kwarg):
    tasks = {
      "vm": loop.create_task(self.__get_attached_devices_vm(account= account, disk= disk, loop= loop)),
      "vmss": loop.create_task(self.__get_attached_devices_vmss(account= account, disk= disk, loop= loop))
    }

    await asyncio.wait(tasks.values())

    return_data = []
    for _, item in tasks.items():
      if item.result() is None:
        continue
      
      return_data += item.result()

    return return_data

  async def __get_account_disk_merged(self, account, cost_by_resource, item, loop, *args, **kwarg):
    return self.get_common().helper_type().dictionary().merge_dictionary([
        {},
        await self.get_base_return_data(
          account= self.get_cloud_client().serialize_resource(resource= account),
          resource_id= self.get_cloud_client().get_resource_id(resource= item),
          resource= item,
          region= self.get_cloud_client().get_resource_location(resource= item),
          resource_groups= [self.get_cloud_client().get_resource_group_from_resource(resource= item)],
        ),         
        {
          "extra_attached_devices": (await self.__get_attached_devices(account=account, disk= item, loop= loop)),
          "extra_resource_cost": {
            "period": cost_by_resource_results.get(self.get_cloud_client().get_resource_id(resource= item)) for period, cost_by_resource_results in cost_by_resource.items()
          }
        }])

  def __process_vmss_vm_disks_skip(self, vm: VirtualMachineScaleSetVM, *args, **kwarg):
    if not hasattr(vm, "storage_profile"):
      return True

    if not hasattr(vm.storage_profile, "os_disk"):
      return True

    if not hasattr(vm.storage_profile.os_disk, "managed_disk"):
      return True

    return False

  def __convert_vmss_vm_disks_disks_creationData(self, vm: VirtualMachineScaleSetVM, *args, **kwarg):
    creationData = {
      "createOption": vm.storage_profile.os_disk.create_option,
      "imageReference": vm.storage_profile.image_reference,
    }
    return creationData

  async def __convert_vmss_vm_disks_disks(self, vmss:VirtualMachineScaleSet, vm: VirtualMachineScaleSetVM, *args, **kwarg):

    return_disks = [
      await self.get_base_return_data(
        resource_id= vm.storage_profile.os_disk.managed_disk.id,
        resource= vm.storage_profile.os_disk,
        region= self.get_cloud_client().get_resource_location(resource= vmss),
        resource_groups= [self.get_cloud_client().get_resource_group_from_resource(resource= vmss)],
      )
    ]
    
    if len(vm.storage_profile.data_disks) > 0:
      for disk in vm.storage_profile.data_disks:
        return_disks.append(
          await self.get_base_return_data(
            resource_id= disk.managed_disk.id,
            resource= disk,
            region= self.get_cloud_client().get_resource_location(resource= vmss),
            resource_groups= [self.get_cloud_client().get_resource_group_from_resource(resource= vmss)],
          )
        )

    return return_disks

  async def __process_vmss_vm_disks(self, vmss:VirtualMachineScaleSet, vm: VirtualMachineScaleSetVM, *args, **kwarg):
    return (await self.__convert_vmss_vm_disks_disks(vm= vm, vmss= vmss))
    
    

  async def __get_account_vmss_disks(self, account, cost_by_resource, loop, *args, **kwarg):
    return_disks = []
    account_id = self.get_cloud_client().get_account_id(account= account)
    if self.__vmss.get(account_id) is None:
      return []
    for vmss in self.__vmss.get(account_id):
      for vm in vmss["extra_vmss_vms"]:
        vm_object = self.get_cloud_client().deserialize_resource(aztype= VirtualMachineScaleSetVM, resource= vm)
        if self.__process_vmss_vm_disks_skip(vm=vm_object):
          continue
        return_disks += (await self.__process_vmss_vm_disks(vm= vm_object, vmss= self.get_cloud_client().deserialize_resource(aztype= VirtualMachineScaleSet, resource= vmss)))
    
    disk_tasks = [ loop.create_task(self.__get_account_disk_merged(account= account, cost_by_resource= cost_by_resource, item= item, loop= loop)) for item in return_disks]
    if len(disk_tasks) < 1:
      return []
    await asyncio.wait(disk_tasks)
    return [ task.result() for task in disk_tasks]
        

  async def __get_account_disks(self, account, cost_by_resource, loop, *args, **kwarg):
    client = ComputeManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))

    disk_tasks = [ loop.create_task(self.__get_account_disk_merged(account= account, cost_by_resource= cost_by_resource, item= item, loop= loop)) for item in self.get_cloud_client().sdk_request(
           tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
           lambda_sdk_command=lambda: client.disks.list()
          )]
    if len(disk_tasks) < 1:
      return []
    await asyncio.wait(disk_tasks)
    return [ task.result() for task in disk_tasks]


  async def _process_account_data(self, account, loop, *args, **kwarg):
    cost_by_resource = {}
    costmanagement_client = CostManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))

    for period, query in self.__cost_by_resource_query_definition.items():
      cost_by_resource[period] = {}
      try:
        resource_cost_details = self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: costmanagement_client.query.usage(
            scope= f'{self.get_cloud_client().get_account_prefix()}{self.get_cloud_client().get_account_id(account= account)}',
            parameters= query,
          )
        )
      except:
        resource_cost_details = None

      if resource_cost_details is not None:
        for row in resource_cost_details.rows:
          if cost_by_resource[period].get(row[1].lower()) is None:
            cost_by_resource[period][row[1].lower()] = row[0]
            continue

          cost_by_resource[period][row[1].lower()] += row[0]
          
    
    return {
      "account": account,
      "data": (
        (await self.__get_account_disks(account= account, cost_by_resource= cost_by_resource, loop= loop)) +
        (await self.__get_account_vmss_disks(account= account, cost_by_resource= cost_by_resource, loop= loop)) 
      )
    }


from threemystic_cloud_data_client.cloud_providers.base_class.base import cloud_data_client_provider_base as base
from threemystic_cloud_client.cloud_client import cloud_client

class cloud_data_client(base): 
  """This is a library to help with the interaction with the cloud providers"""

  def __init__(self, logger = None, common = None, *args, **kwargs) -> None: 
    if "provider" not in kwargs:
      kwargs["provider"] = ""
    super().__init__(common= common, logger_name= "cloud_data_client", logger= logger, *args, **kwargs)
    
  def version(self, *args, **kwargs):
    if hasattr(self, "_version"):
      return self._version
    import threemystic_cloud_data_client.__version__ as __version__
    self._version = __version__.__version__
    return self.version()
    
  def get_supported_providers(self, *args, **kwargs):
    return super().get_supported_providers()
  
  def init_client(self, provider, refresh = False, *args, **kwargs):
    provider = self.get_common().helper_type().string().set_case(string_value= provider, case= "lower") if provider is not None else ""

    if provider not in self.get_supported_providers():
      raise self.get_common().exception().exception(
        exception_type = "argument"
      ).not_implemented(
        logger= self.get_logger(),
        name = "provider",
        message = f"Unknown Cloud Provided: {provider}.\nSupported Cloud Providers{self.get_supported_providers()}"
      )

    if not hasattr(self, "_client"):
      self._client = {}

    if self._client.get(provider) is not None and not refresh:
      return

    if provider == "azure":
      from threemystic_cloud_data_client.cloud_providers.azure.client import cloud_data_client_azure_client as provider_cloud_data_client
      self._client[provider] = provider_cloud_data_client(
        cloud_data_client= self,
        cloud_client = cloud_client(
          logger= self.get_logger(), 
          common= self.get_common()
        ).client(provider= provider),
        *args, **kwargs
      )
      return
    
    if provider == "aws":
      from threemystic_cloud_data_client.cloud_providers.aws.client import cloud_data_client_aws_client as provider_cloud_data_client
      self._client[provider] = provider_cloud_data_client(
        cloud_data_client= self,
        cloud_client = cloud_client(
          logger= self.get_common().get_logger(), 
          common= self.get_common()
        ).client(provider= provider),
        *args, **kwargs
      )
      return  
       
    raise self.get_common().exception().exception(
      exception_type = "argument"
    ).not_implemented(
      logger= self.get_common().get_logger(), 
      name = "provider",
      message = f"Unknown Cloud Provided: {provider}.\nSupported Cloud Providers{self.get_supported_providers()}"
    )

  def client(self, provider = None, *args, **kwargs):
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= provider):
      provider = self.get_provider()
      if self.get_common().helper_type().string().is_null_or_whitespace(string_value= provider):
        provider = self.get_default_provider()
        if self.get_common().helper_type().string().is_null_or_whitespace(string_value= provider):
          raise self.get_common().exception().exception(
            exception_type = "argument"
          ).not_implemented(
            logger= self.get_common().get_logger(), 
            name = "provider",
            message = f"provider cannot be null or whitespace"
          )
  
    provider = self.get_common().helper_type().string().set_case(string_value= provider, case= "lower")
    if hasattr(self, "_client"):
      if self._client.get(provider) is not None:
        return self._client.get(provider)
    
    self.init_client(provider= provider,  *args, **kwargs)
    return self.client(provider= provider, *args, **kwargs)
    
    

  


from threemystic_cloud_data_client.cloud_providers.azure.client.actions.base_class.base import cloud_data_client_azure_client_action_base as base
import asyncio
from azure.mgmt.cosmosdb import CosmosDBManagementClient
from azure.mgmt.sql import SqlManagementClient
from azure.mgmt.sqlvirtualmachine import SqlVirtualMachineManagementClient
from azure.mgmt.rdbms.mysql import MySQLManagementClient
from azure.mgmt.rdbms.mariadb import MariaDBManagementClient
from azure.mgmt.rdbms.postgresql import PostgreSQLManagementClient

class cloud_data_client_azure_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="database", 
      logger_name= "cloud_data_client_azure_client_action_database",
      *args, **kwargs)
  
  async def __process_get_db_sql(self, client, account, *args, **kwargs):    
    try:
      return_data = []
      for db in self.get_cloud_client().sdk_request(
        tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
        lambda_sdk_command=lambda: client.managed_instances.list()
        ):
        return_data.append(
          {"extra":{
            "extra_dbtype": "sqlmi"
          }, "resource": db 
        })        
      
      for db in self.get_cloud_client().sdk_request(
        tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
        lambda_sdk_command=lambda: client.servers.list()
        ):
        return_data.append({
          "extra":{
            "extra_dbtype": "sql",
            "extra_databases": [ self.get_cloud_client().serialize_resource(resource= pool) for pool in self.get_cloud_client().sdk_request(
              tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
              lambda_sdk_command=lambda: client.databases.list_by_server(resource_group_name= self.get_cloud_client().get_resource_group_from_resource(resource= db), server_name= db.name)
              )
            ],
            "extra_elastic_pool": [ 
              self.get_common().helper_type().dictionary().merge_dictionary([
                {},
                {
                  "extra_databases": [ self.get_cloud_client().serialize_resource(resource= pool) for pool in self.get_cloud_client().sdk_request(
                    tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
                    lambda_sdk_command=lambda: client.databases.list_by_elastic_pool(resource_group_name= self.get_cloud_client().get_resource_group_from_resource(resource= db), server_name= db.name, elastic_pool_name= pool.name)
                    )
                  ]
                },
                self.get_cloud_client().serialize_resource(resource= pool)]) for pool in self.get_cloud_client().sdk_request(
              tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
              lambda_sdk_command=lambda: client.elastic_pools.list_by_server(resource_group_name= self.get_cloud_client().get_resource_group_from_resource(resource= db), server_name= db.name)
              )
            ]
          }, 
          "resource": db 
        })

        return return_data



    except Exception as err:
      self.get_common().get_logger().exception(
        msg= f"__process_get_db_sql: {err}",
        extra={
          "exception": err
        }
      )
      return []
  
  async def __process_get_db_sqlvm(self, client, account, *args, **kwargs):    
    try:

      return_data = []
      for db in self.get_cloud_client().sdk_request(
        tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
        lambda_sdk_command=lambda: client.sql_virtual_machines.list()
        ):
        return_data.append(
          {"extra":{
            "extra_dbtype": "sqlvm"
          }, "resource": db 
        }) 
      
      return return_data


    except Exception as err:
      self.get_common().get_logger().exception(
        msg= f"__process_get_db_sqlvm: {err}",
        extra={
          "exception": err
        }
      )
      return []
  
  async def __process_get_db_mysql(self, client, account, *args, **kwargs):    
    try:

      return_data = []
      for db in self.get_cloud_client().sdk_request(
        tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
        lambda_sdk_command=lambda: client.servers.list()
        ):
        return_data.append(
          {"extra":{
            "extra_dbtype": "mysql",
            "extra_databases": [ self.get_cloud_client().serialize_resource(resource= pool) for pool in self.get_cloud_client().sdk_request(
              tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
              lambda_sdk_command=lambda: client.databases.list_by_server(resource_group_name= self.get_cloud_client().get_resource_group_from_resource(resource= db), server_name= db.name)
              )
            ]
          }, "resource": db 
        }) 
      
      return return_data


    except Exception as err:
      self.get_common().get_logger().exception(
        msg= f"__process_get_db_mysql: {err}",
        extra={
          "exception": err
        }
      )
      return []
  
  async def __process_get_db_postgres(self, client, account, *args, **kwargs):    
    try:

      return_data = []
      for db in self.get_cloud_client().sdk_request(
        tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
        lambda_sdk_command=lambda: client.sql_virtual_machines.list()
        ):
        return_data.append(
          {"extra":{
            "extra_dbtype": "postgres",
            "extra_databases": [ self.get_cloud_client().serialize_resource(resource= pool) for pool in self.get_cloud_client().sdk_request(
              tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
              lambda_sdk_command=lambda: client.databases.list_by_server(resource_group_name= self.get_cloud_client().get_resource_group_from_resource(resource= db), server_name= db.name)
              )
            ]
          }, "resource": db 
        }) 
      
      return return_data


    except Exception as err:
      self.get_common().get_logger().exception(
        msg= f"__process_get_db_postgres: {err}",
        extra={
          "exception": err
        }
      )
      return []
  
  async def __process_get_db_mariadb(self, client, account, *args, **kwargs):    
    try:

      return_data = []
      for db in self.get_cloud_client().sdk_request(
        tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
        lambda_sdk_command=lambda: client.sql_virtual_machines.list()
        ):
        return_data.append(
          {"extra":{
            "extra_dbtype": "mariadb",
            "extra_databases": [ self.get_cloud_client().serialize_resource(resource= pool) for pool in self.get_cloud_client().sdk_request(
              tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
              lambda_sdk_command=lambda: client.databases.list_by_server(resource_group_name= self.get_cloud_client().get_resource_group_from_resource(resource= db), server_name= db.name)
              )
            ]
          }, "resource": db 
        }) 
      
      return return_data


    except Exception as err:
      self.get_common().get_logger().exception(
        msg= f"__process_get_db_maria: {err}",
        extra={
          "exception": err
        }
      )
      return []
    
  async def _process_account_data(self, account, loop, *args, **kwargs):
    # database categories
    # https://azure.microsoft.com/en-us/products/category/databases/

    sql_client = SqlManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    sqlvm_client = SqlVirtualMachineManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    mysql_client = MySQLManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    mariadb_client = MariaDBManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    postgres_client = PostgreSQLManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    for test in sql_client.servers.list( ):
      test.ta
    tasks = {
       "sql_client": loop.create_task(self.__process_get_db_sql(client= sql_client,account= account)),
       "sqlvm_client": loop.create_task(self.__process_get_db_sqlvm(client= sqlvm_client,account= account)),
       "mysql_client": loop.create_task(self.__process_get_db_mysql(client= mysql_client,account= account)),
       "mariadb_client": loop.create_task(self.__process_get_db_mariadb(client= mariadb_client,account= account)),
       "postgres_client": loop.create_task(self.__process_get_db_postgres(client= postgres_client,account= account)),
    }

    await asyncio.wait(tasks.values())
    return_data = {
        "account": account,
        "data": []
    }

    for db_data in tasks.values():
      if db_data.result() is None:
        continue
      for item in db_data.result():
        return_data["data"].append( self.get_common().helper_type().dictionary().merge_dictionary([
            {},
            await self.get_base_return_data(
              account= self.get_cloud_client().serialize_resource(resource= account),
              resource_id= self.get_cloud_client().get_resource_id(resource= item.get("resource")),
              resource= item.get("resource"),
              region= self.get_cloud_client().get_resource_location(resource= item.get("resource")),
              resource_groups= [self.get_cloud_client().get_resource_group_from_resource(resource= item.get("resource"))],
            ),
            item.get("extra"),
          ]))

    return return_data



from threemystic_cloud_data_client.cloud_providers.base_class.base import cloud_data_client_provider_base as base

class cloud_data_client_provider_azure_base(base):
  def __init__(self, *args, **kwargs):
    super().__init__(provider= "azure", *args, **kwargs)  



from threemystic_cloud_data_client.cloud_providers.base_class.base_client import cloud_data_client_provider_base_client as base

class cloud_data_client_aws_client(base):
  def __init__(self, *args, **kwargs):
    super().__init__(provider= "aws", logger_name= "cloud_data_client_aws_client", *args, **kwargs)
    
    
  def get_client_parser_args_actions(self, *args, **kwargs):
    return {
       "--elastisearch,--es": {
        "default": None, 
        "const": "elastisearch",
        "dest": "data_action",
        "help": "Data Action: This pulls Elasticsearch data from aws",
        "action": 'store_const' # could look into append_const
      },
    }
  


from threemystic_cloud_data_client.cloud_providers.azure.client.actions.base_class.base import cloud_data_client_azure_client_action_base as base
import asyncio
from azure.mgmt.keyvault import KeyVaultManagementClient
from azure.keyvault.secrets import SecretClient
from azure.keyvault.keys import KeyClient


class cloud_data_client_azure_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="certificates", 
      logger_name= "cloud_data_client_azure_client_action_certificates",
      *args, **kwargs)
  
  async def __process_get_resources_key_vault_keys(self, account, key_vault, *args, **kwargs):    
    try:
      client = KeyClient(vault_url= f'https://{key_vault}.vault.azure.net/', credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
      
      return [ key for key in self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: client.list_properties_of_keys()
        )
      ]
           
    except Exception as err:
      return []
    
  async def __process_get_resources_key_vault_secrets(self, account, key_vault, *args, **kwargs):    
    try:
      client = SecretClient(vault_url= f'https://{key_vault}.vault.azure.net/', credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    

      return [ secret for secret in self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: client.list_properties_of_secrets()
        )
      ]
           
    except Exception as err:
      return []
    
  async def __process_get_resources_key_vaults(self, account, *args, **kwargs):    
    try:
      client = KeyVaultManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    

      return { self.get_cloud_client().get_resource_id(resource= kv):kv for kv in self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: client.vaults.list()
        )
      }
           
    except Exception as err:
      return {} 

  
  async def _process_account_data(self, account, loop, *args, **kwargs):
    
    keyvaults = await self.__process_get_resources_key_vaults(account= account)

    for kv in keyvaults.values():
      print(self.get_cloud_client().serialize_resource(resource= kv))
      break

    return {
        "account": account,
        "data": [ 
          # self.get_common().helper_type().dictionary().merge_dictionary([
          #   {},
          #   await self.get_base_return_data(
          #     account= self.get_cloud_client().serialize_resource(resource= account),
          #     resource_id= self.get_cloud_client().get_resource_id(resource= item),
          #     resource= item,
          #     region= self.get_cloud_client().get_resource_location(resource= item),
          #     resource_groups= [self.get_cloud_client().get_resource_group_from_resource(resource= item)],
          #   ),
          #   {
          #     "extra_resource": self.get_cloud_client().serialize_resource(tasks["resource"].result().get(self.get_cloud_client().get_resource_id(resource= item))),
          #     "extra_availability_set": tasks["availability_sets"].result().get(self.get_cloud_client().get_resource_id(resource= item)),
          #     "extra_nics": tasks["nics"].result().get(self.get_cloud_client().get_resource_id(resource= item)),
          #     "extra_load_balancers": await self._process_account_data_get_vm_load_balancers(
          #       vm_nics= tasks["nics"].result().get(self.get_cloud_client().get_resource_id(resource= item)),
          #       load_balancers_by_nics = tasks["load_balancers"].result()
          #     ),
          #   },
          # ]) for item in self.get_cloud_client().sdk_request(
          #  tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          #  lambda_sdk_command=lambda: client.virtual_machines.list_all()
          # )
        ]
    }


from threemystic_common.base_class.base_provider import base
from threemystic_common.base_class.base_script_options import base_process_options
import textwrap, argparse
import asyncio


class cloud_data_client_provider_base_client(base):
  def __init__(self, force_action_arguments = None, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.__set_cloud_client(*args, **kwargs)
    self.__set_cloud_data_client(*args, **kwargs)
    self.__set_suppres_parser_help(*args, **kwargs)
    
    self._default_parser_init = {
      "prog": f'3mystic_cloud_data_client -d -p {kwargs["provider"]}',
      "formatter_class": argparse.RawDescriptionHelpFormatter,
      "description": textwrap.dedent('''\
      Requires additional settings.
        One data Action (if more than one is selected only the last one will be ran)
      '''),
      "add_help": False,
      "epilog": ""
    }
  
    self._set_arguments_from_parameters(force_action_arguments= force_action_arguments, *args, **kwargs)
    
    self._set_data_action()
  
  def get_client_parser_args_actions(self, *args, **kwargs):
    return {}

  def get_default_parser_args_actions(self, *args, **kwargs):
    return {
      "--cloudstorage": {
        "default": None, 
        "const": "cloudstorage",
        "dest": "data_action",
        "help": "Data Action: This pulls Cloud Storage (S3/Storage Accounts) for the provider",
        "action": 'store_const' # could look into append_const
      },
      "--budget": {
        "default": None, 
        "const": "budget",
        "dest": "data_action",
        "help": "Data Action: This pulls a general budget to provide you insights in your accounts/subscriptions",
        "action": 'store_const'
      },
      # "--secrets": {
      #   "default": None, 
      #   "const": "secrets",
      #   "dest": "data_action",
      #   "help": "Data Action: Pulls Key Vaults Secrets and Keys / Paramater Store",
      #   "action": 'store_const'
      # },
      "--storage": {
        "default": None, 
        "const": "storage",
        "dest": "data_action",
        "help": "Data Action: This pulls either VM Disks or EC2 Storage depending on the provider",
        "action": 'store_const'
      },
      "--vm": {
        "default": None, 
        "const": "vm",
        "dest": "data_action",
        "help": "Data Action: This pulls either EC2 or VM depending on the provider",
        "action": 'store_const'
      },
      "--vmss": {
        "default": None, 
        "const": "vmss",
        "dest": "data_action",
        "help": "Data Action: This pulls either ASG or VMSS depending on the provider",
        "action": 'store_const'
      },
      "--certificates,--ssl": {
        "default": None, 
        "const": "certificates",
        "dest": "data_action",
        "help": "Data Action: This pulls Certificats information from services like acm/keyvault",
        "action": 'store_const'
      },
      "--dns": {
        "default": None, 
        "const": "dns",
        "dest": "data_action",
        "help": "Data Action: This pulls DNS Data (private DNS/Public/Route53)",
        "action": 'store_const'
      },
      "--datawarehouse": {
        "default": None, 
        "const": "datawarehouse",
        "dest": "data_action",
        "help": "Data Action: This pulls Data Warehouse (Synapse/RedShift)",
        "action": 'store_const'
      },
      "--db,--database": {
        "default": None, 
        "const": "database",
        "dest": "data_action",
        "help": "Data Action: This pulls the various non-inmemory databses",
        "action": 'store_const'
      },
      "--clouddb,--clouddatabase,--cdb": {
        "default": None, 
        "const": "clouddb",
        "dest": "data_action",
        "help": "Data Action: This pulls the special cloud DB ie Azure/CosmosDB, AWS/DynamoDB",
        "action": 'store_const'
      },
      "--memorydb": {
        "default": None, 
        "const": "memorydb",
        "dest": "data_action",
        "help": "Data Action: This pulls memorydbs IE Redis/elasticache",
        "action": 'store_const'
      },
      "--vmimage": {
        "default": None, 
        "const": "vmimage",
        "dest": "data_action",
        "help": "Data Action: This pulls either Image Galleries or all the AMis",
        "action": 'store_const'
      },
    }
  
  def get_parser_args(self, *args, **kwargs):
    if hasattr(self, "_data_parser_args"):
      return self._data_parser_args
    
    
    self._data_parser_args = self.get_common().helper_type().dictionary().merge_dictionary([
      {},
      {
        "--filter,-f": {
          "default": None,
          "type": str,
          "dest": "data_filter",
          "help": "Filter: JSON string representing the filter data. ex {\"condition\":\"and\", \"filters\":[{\"condition\":\"iequals\",\"key\": [\"properties\",\"storageProfile\",\"osDisk\",\"osType\"], \"value\": \"windows\"}]}",
          "action": 'store'
        },
        "--hideempty, --hide-empty": {
          "default": False,
          "dest": "data_hideempty",
          "help": "Hide accounts with no data. Default False",
          "action": 'store_true'
        },
        "--account": {
          "default": None,
          "type": str,
          "dest": "data_accounts",
          "help": "Filter: A comma seperated list of accounts to filter, put a minus to exclude accounts. 123454,12345,-234534",
          "action": 'store'
        }
      },
      self.get_default_parser_args_actions(),
      self.get_client_parser_args_actions()
    ])
    return self.get_parser_args()
  
  def get_suppres_parser_help(self, *args, **kwargs):
    if hasattr(self, "_suppres_parser_help"):
      return self._suppres_parser_help
    return False
    

  def __set_suppres_parser_help(self, suppress_parser_help = False, *args, **kwargs):
    self._suppres_parser_help = suppress_parser_help

  def __get_action_parser_options(self, *args, **kwargs):
    if hasattr(self, "_action_process_options"):
      return self._action_process_options
    
    self._action_process_options = base_process_options(common= self.get_common())
    return self.__get_action_parser_options()
    
  def _get_action_parser(self, *args, **kwargs):
    if hasattr(self, "_action_parser"):
      return self._action_parser
    
    
    self._action_parser = self.__get_action_parser_options().get_parser(
      parser_init_kwargs = self._default_parser_init,
      parser_args = self.get_common().helper_type().dictionary().merge_dictionary([
        {},
        self.get_parser_args(),
      ])
    )
    return self._get_action_parser()
  
  def get_data_run_key(self, data_key = None, default_value = {}, *args, **kwargs):
    if hasattr(self, "_data_arg_param_values"):
      if data_key is None:
        return self._data_arg_param_values
      if data_key in self._data_arg_param_values:
        return self._data_arg_param_values.get(data_key)
    
    return default_value
  
  def get_data_run_action(self, *args, **kwargs):
    if hasattr(self, "_data_arg_param_values"):
      return self._data_arg_param_values.get("data_action")
    
    return None
  
  def _process_data_run_filter(self, data_filter, *args, **kwargs):
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= data_filter):
      data_filter = {}
    
    if data_filter is None:
      data_filter = {}
    
    if self.get_common().helper_type().general().is_type(obj= data_filter, type_check= str):
      try:
        data_filter = self.get_common().helper_json().loads(data= data_filter)
      except Exception as err:
        # will look to do something for various verbocity levels
        # print(err) 
        pass
    
    if not self.get_common().helper_type().general().is_type(obj= data_filter, type_check= dict):
      return {}
    
    return data_filter    
  
  
  def _set_arguments_from_parameters(self, data_action = None, data_filter = None, data_hideempty = None, data_accounts = None, *args, **kwargs):
    """
    If the values for the data_* params are not either None or an empty string it will override whatever is in the arguments from the script call
    """
    processed_info = self.__get_action_parser_options().process_opts(
      parser = self._get_action_parser()
    )

    self._data_arg_param_values={
      "data_action": data_action,
      "data_filter": data_filter,
      "data_hideempty": data_hideempty,
      "data_accounts": data_accounts,
    }

    for key, item in self._data_arg_param_values.items():
      if item is None or self.get_common().helper_type().string().is_null_or_whitespace(string_value= item):
        self._data_arg_param_values[key] = processed_info["processed_data"].get(key)


    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= self._data_arg_param_values.get("data_action")):
      if not self.get_suppres_parser_help():
        self._get_action_parser().print_help()
        return None
      
    self._data_arg_param_values["data_filter"] = self._process_data_run_filter(data_filter= data_filter)
    self._data_arg_param_values["data_hideempty"] = self.get_common().helper_type().bool().is_true(check_value= self._data_arg_param_values["data_hideempty"])
    

  def get_data_action(self, action = None, *args, **kwargs):
    if hasattr(self, "_data_action_data"):
      
      if self.get_common().helper_type().string().is_null_or_whitespace(string_value= action):
        action = self._data_action_data["default"]
      
      if self._data_action_data.get(action) is not None:
        return self._data_action_data.get(action)
      
      self._data_action_data[action] = self._process_data_action(
        provider = self.get_provider(),
        action= action, 
        *args, **kwargs)
      return self.get_data_action(action= action, *args, **kwargs)
  
  def _set_data_action(self, *args, **kwargs):
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= self.get_data_run_action()):
      return

    try:      
      action = self.get_common().helper_type().string().set_case(string_value= self.get_data_run_action() , case= "lower")
      if action != "all":
        self._data_action_data = {
          "default": action,
          action: self._process_data_action(
          provider = self.get_provider(),
          action= action, 
          *args, **kwargs)
        }
        return
      
      self._data_action_data = {
        arg_key: self._process_data_action(
          provider = self.get_provider(),
          action= arg.get("const"), 
          *args, **kwargs)
        for arg_key, arg in self.get_default_parser_args_actions().items() if arg.get("const") != "all"
      }
      self._data_action_data["default"] = list(self._data_action_data.keys())[0]
      
    except Exception as err:
      self.get_common().get_logger().exception(f"The action {self.get_data_run_action()} is unknown", extra={"exception": err})
      if not self.get_suppres_parser_help():
        self._get_action_parser().print_help()
  
  def _process_data_action(self, provider, action, *args, **kwargs):    
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= action):
      return None
    data_action = __import__(f'threemystic_cloud_data_client.cloud_providers.{provider}.client.actions.{action}', fromlist=[f'cloud_data_client_{provider}_client_action'])
    process_data_action = getattr(data_action, f'cloud_data_client_{provider}_client_action')(
      cloud_data_client= self,
      common= self.get_common(),
      logger= self.get_common().get_logger()
    )
    return process_data_action
  
  def run(self, *args, **kwargs):
    if self.get_data_action() is None:
      return 

    results = asyncio.run(self.get_data_action().main(run_params= self.get_data_run_key(data_key= None) ))
    self.get_data_action().format_results(results= results, output_format= self.get_cloud_data_client().get_default_output_format())
  
  def get_cloud_client(self, *args, **kwargs):
    return self.__cloud_client
  
  def __set_cloud_client(self, cloud_client, *args, **kwargs):
    self.__cloud_client = cloud_client
  
  def get_cloud_data_client(self, *args, **kwargs):
    return self.__cloud_data_client
  
  def __set_cloud_data_client(self, cloud_data_client, *args, **kwargs):
    self.__cloud_data_client = cloud_data_client
    
  
  
    

  
  

